from time import time
from typing import Generator

from httpx import Request, Response

from firebolt.client.auth.base import Auth
from firebolt.client.constants import _REQUEST_ERRORS
from firebolt.utils.exception import AuthenticationError
from firebolt.utils.usage_tracker import get_user_agent_header


class _RequestBasedAuth(Auth):
    """Base abstract class for http request based authentication."""

    def __init__(self, use_token_cache: bool = True):
        self._user_agent = get_user_agent_header()
        super().__init__(use_token_cache)

    def _make_auth_request(self) -> Request:
        """Create an HTTP request required for authentication.
        Returns:
            Request: HTTP request, required for authentication.
        """
        raise NotImplementedError()

    @staticmethod
    def _check_response_error(response: dict) -> None:
        """Check if response data contains errors.
        Args:
            response (dict): Response data
        Raises:
            AuthenticationError: Were unable to authenticate
        """
        if "error" in response:
            raise AuthenticationError(
                response.get("message", "unknown server error"),
            )

    def get_new_token_generator(self) -> Generator[Request, Response, None]:
        """Get new token using username and password.
        Yields:
            Request: An http request to get token. Expects Response to be sent back
        Raises:
            AuthenticationError: Error while authenticating with provided credentials
        """
        try:
            response = yield self._make_auth_request()
            response.raise_for_status()

            parsed = response.json()
            self._check_response_error(parsed)

            self._token = parsed["access_token"]
            self._expires = int(time()) + int(parsed["expires_in"])

        except _REQUEST_ERRORS as e:
            raise AuthenticationError(repr(e)) from e



import json
from typing import Any

from pydantic import BaseModel


class FireboltBaseModel(BaseModel):
    class Config:
        allow_population_by_field_name = True
        extra = "forbid"

    def jsonable_dict(self, *args: Any, **kwargs: Any) -> dict:
        """
        Generate a dictionary representation of the service that contains serialized
        primitive types, and is therefore JSON-ready.

        This could be replaced with something native once this issue is resolved:
        https://github.com/samuelcolvin/pydantic/issues/1409

        This function is intended to improve the compatibility with HTTPX, which
        expects to take in a dictionary of primitives as input to the JSON parameter
        of its request function. See: https://www.python-httpx.org/api/#helper-functions
        """
        return json.loads(self.json(*args, **kwargs))



from __future__ import annotations

from functools import wraps
from threading import Lock
from typing import Any, Generator, List, Optional, Sequence, Tuple, Union

from readerwriterlock.rwlock import RWLockWrite

from firebolt.async_db._types import ColType
from firebolt.async_db.cursor import BaseCursor as AsyncBaseCursor
from firebolt.async_db.cursor import (
    ParameterType,
    QueryStatus,
    check_not_closed,
    check_query_executed,
)
from firebolt.utils.util import async_to_sync


class Cursor(AsyncBaseCursor):
    """
    Class, responsible for executing queries to Firebolt Database.
    Should not be created directly,
    use :py:func:`connection.cursor <firebolt.async_db.connection.Connection>`

    Args:
        description: Information about a single result row
        rowcount: The number of rows produced by last query
        closed: True if connection is closed, False otherwise
        arraysize: Read/Write, specifies the number of rows to fetch at a time
            with the :py:func:`fetchmany` method
    """

    __slots__ = AsyncBaseCursor.__slots__ + (
        "_query_lock",
        "_idx_lock",
    )

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._query_lock = RWLockWrite()
        self._idx_lock = Lock()
        super().__init__(*args, **kwargs)

    @wraps(AsyncBaseCursor.execute)
    def execute(
        self,
        query: str,
        parameters: Optional[Sequence[ParameterType]] = None,
        skip_parsing: bool = False,
        async_execution: Optional[bool] = False,
    ) -> Union[int, str]:
        with self._query_lock.gen_wlock():
            return async_to_sync(super().execute)(
                query, parameters, skip_parsing, async_execution
            )

    @wraps(AsyncBaseCursor.executemany)
    def executemany(
        self,
        query: str,
        parameters_seq: Sequence[Sequence[ParameterType]],
        async_execution: Optional[bool] = False,
    ) -> Union[int, str]:
        with self._query_lock.gen_wlock():
            return async_to_sync(super().executemany)(
                query, parameters_seq, async_execution
            )

    @wraps(AsyncBaseCursor._get_next_range)
    def _get_next_range(self, size: int) -> Tuple[int, int]:
        with self._idx_lock:
            return super()._get_next_range(size)

    @wraps(AsyncBaseCursor.fetchone)
    def fetchone(self) -> Optional[List[ColType]]:
        with self._query_lock.gen_rlock():
            return super().fetchone()

    @wraps(AsyncBaseCursor.fetchmany)
    def fetchmany(self, size: Optional[int] = None) -> List[List[ColType]]:
        with self._query_lock.gen_rlock():
            return super().fetchmany(size)

    @wraps(AsyncBaseCursor.fetchall)
    def fetchall(self) -> List[List[ColType]]:
        with self._query_lock.gen_rlock():
            return super().fetchall()

    @wraps(AsyncBaseCursor.nextset)
    def nextset(self) -> None:
        with self._query_lock.gen_rlock(), self._idx_lock:
            return super().nextset()

    # Iteration support
    @check_not_closed
    @check_query_executed
    def __iter__(self) -> Generator[List[ColType], None, None]:
        while True:
            row = self.fetchone()
            if row is None:
                return
            yield row

    @wraps(AsyncBaseCursor.get_status)
    def get_status(self, query_id: str) -> QueryStatus:
        with self._query_lock.gen_rlock():
            return async_to_sync(super().get_status)(query_id)

    @wraps(AsyncBaseCursor.cancel)
    def cancel(self, query_id: str) -> None:
        with self._query_lock.gen_rlock():
            return async_to_sync(super().cancel)(query_id)



from typing import Optional

from firebolt.client.auth.base import AuthRequest
from firebolt.client.auth.request_auth_base import _RequestBasedAuth
from firebolt.utils.token_storage import TokenSecureStorage
from firebolt.utils.urls import AUTH_SERVICE_ACCOUNT_URL
from firebolt.utils.util import cached_property


class ServiceAccount(_RequestBasedAuth):
    """Service Account authentication class for Firebolt Database.

    Gets authentication token using
    provided credentials and updates it when it expires.

    Args:
        client_id (str): Client ID
        client_secret (str): Client secret
        use_token_cache (bool): True if token should be cached in filesystem;
            False otherwise

    Attributes:
        client_id (str): Client ID
        client_secret (str): Client secret
    """

    __slots__ = (
        "client_id",
        "client_secret",
        "_token",
        "_expires",
        "_use_token_cache",
        "_user_agent",
    )

    requires_response_body = True

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        use_token_cache: bool = True,
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        super().__init__(use_token_cache)

    def copy(self) -> "ServiceAccount":
        """Make another auth object with same credentials.

        Returns:
            ServiceAccount: Auth object
        """
        return ServiceAccount(self.client_id, self.client_secret, self._use_token_cache)

    @cached_property
    def _token_storage(self) -> Optional[TokenSecureStorage]:
        """Token filesystem cache storage.

        This is evaluated lazily, only if caching is enabled

        Returns:
            TokenSecureStorage: Token filesystem cache storage
        """
        return TokenSecureStorage(username=self.client_id, password=self.client_secret)

    def _make_auth_request(self) -> AuthRequest:
        """Get new token using username and password.

        Yields:
            Request: An http request to get token. Expects Response to be sent back

        Raises:
            AuthenticationError: Error while authenticating with provided credentials
        """

        response = self.request_class(
            "POST",
            AUTH_SERVICE_ACCOUNT_URL,
            headers={
                "User-Agent": self._user_agent,
            },
            data={
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "grant_type": "client_credentials",
            },
        )
        return response



from typing import Any, Optional

from anyio._core._eventloop import get_asynclib
from httpx import URL
from httpx import AsyncClient as HttpxAsyncClient
from httpx import Client as HttpxClient
from httpx import Request, Response
from httpx import codes as HttpxCodes
from httpx._types import AuthTypes

from firebolt.client.auth import Auth
from firebolt.client.auth.base import AuthRequest
from firebolt.client.constants import DEFAULT_API_URL
from firebolt.utils.exception import AccountNotFoundError
from firebolt.utils.urls import ACCOUNT_BY_NAME_URL, ACCOUNT_URL
from firebolt.utils.util import (
    cached_property,
    fix_url_schema,
    merge_urls,
    mixin_for,
)

# Explicitly import all available backend not get into
# anyio race condition during backend import
for backend in ("asyncio", "trio"):
    try:
        get_asynclib(backend)
    except ModuleNotFoundError:
        # Not all backends might be installed
        pass

FireboltClientMixinBase = mixin_for(HttpxClient)  # type: Any


class FireboltClientMixin(FireboltClientMixinBase):
    """HttpxAsyncClient mixin with Firebolt authentication functionality."""

    def __init__(
        self,
        *args: Any,
        account_name: Optional[str] = None,
        api_endpoint: str = DEFAULT_API_URL,
        auth: Optional[Auth] = None,
        **kwargs: Any,
    ):
        self.account_name = account_name
        self._api_endpoint = URL(fix_url_schema(api_endpoint))
        super().__init__(*args, auth=auth, **kwargs)

    def _build_auth(self, auth: Optional[AuthTypes]) -> Optional[Auth]:
        """Create Auth object based on auth provided.

        Overrides ``httpx.Client._build_auth``

        Args:
            auth (AuthTypes): Provided auth

        Returns:
            Optional[Auth]: Auth object

        Raises:
            TypeError: Auth argument has unsupported type
        """
        if auth is None or isinstance(auth, Auth):
            return auth
        raise TypeError(f'Invalid "auth" argument: {auth!r}')

    def _merge_auth_request(self, request: Request) -> Request:
        if isinstance(request, AuthRequest):
            request.url = merge_urls(self._api_endpoint, request.url)
            request._prepare(dict(request.headers))
        return request


class Client(FireboltClientMixin, HttpxClient):
    """An HTTP client, based on httpx.Client.

    Handles the authentication for Firebolt database.
    Authentication can be passed through auth keyword as a tuple or as a
    FireboltAuth instance
    """

    @cached_property
    def account_id(self) -> str:
        """User account ID.

        If account_name was provided during Client construction, returns its ID;
        gets default account otherwise.

        Returns:
            str: Account ID

        Raises:
            AccountNotFoundError: No account found with provided name
        """
        if self.account_name:
            response = self.get(
                url=ACCOUNT_BY_NAME_URL, params={"account_name": self.account_name}
            )
            if response.status_code == HttpxCodes.NOT_FOUND:
                raise AccountNotFoundError(self.account_name)
            # process all other status codes
            response.raise_for_status()
            return response.json()["account_id"]

        # account_name isn't set, use the default account.
        return self.get(url=ACCOUNT_URL).json()["account"]["id"]

    def _send_handling_redirects(
        self, request: Request, *args: Any, **kwargs: Any
    ) -> Response:
        return super()._send_handling_redirects(
            self._merge_auth_request(request), *args, **kwargs
        )


class AsyncClient(FireboltClientMixin, HttpxAsyncClient):
    """An HTTP client, based on httpx.AsyncClient.

    Asynchronously handles authentication for Firebolt database.
    Authentication can be passed through auth keyword as a tuple, or as a
    FireboltAuth instance.
    """

    @cached_property
    async def account_id(self) -> str:
        """User account id.

        If account_name was provided during AsyncClient construction, returns its ID;
        gets default account otherwise.

        Returns:
            str: Account ID

        Raises:
            AccountNotFoundError: No account found with provided name
        """
        if self.account_name:
            response = await self.get(
                url=ACCOUNT_BY_NAME_URL, params={"account_name": self.account_name}
            )
            if response.status_code == HttpxCodes.NOT_FOUND:
                raise AccountNotFoundError(self.account_name)
            # process all other status codes
            response.raise_for_status()
            return response.json()["account_id"]

        # account_name isn't set; use the default account.
        return (await self.get(url=ACCOUNT_URL)).json()["account"]["id"]

    async def _send_handling_redirects(
        self, request: Request, *args: Any, **kwargs: Any
    ) -> Response:
        return await super()._send_handling_redirects(
            self._merge_auth_request(request), *args, **kwargs
        )



from typing import Dict, List

from firebolt.model.region import Region, RegionKey
from firebolt.service.base import BaseService
from firebolt.service.manager import ResourceManager
from firebolt.utils.urls import REGIONS_URL
from firebolt.utils.util import cached_property


class RegionService(BaseService):
    def __init__(self, resource_manager: ResourceManager):
        """
        Service to manage AWS regions (us-east-1, etc)

        Args:
            resource_manager: Resource manager to use
        """

        super().__init__(resource_manager=resource_manager)

    @cached_property
    def regions(self) -> List[Region]:
        """List of available AWS regions on Firebolt."""

        response = self.client.get(url=REGIONS_URL, params={"page.first": 5000})
        return [Region.parse_obj(i["node"]) for i in response.json()["edges"]]

    @cached_property
    def regions_by_name(self) -> Dict[str, Region]:
        """Dict of {RegionLookup to Region}"""

        return {r.name: r for r in self.regions}

    @cached_property
    def regions_by_key(self) -> Dict[RegionKey, Region]:
        """Dict of {RegionKey to Region}"""

        return {r.key: r for r in self.regions}

    @cached_property
    def default_region(self) -> Region:
        """Default AWS region, could be provided from environment."""

        if not self.settings.default_region:
            raise ValueError(
                "The environment variable FIREBOLT_DEFAULT_REGION must be set."
            )
        return self.get_by_name(name=self.settings.default_region)

    def get_by_name(self, name: str) -> Region:
        """Get an AWS region by its name (eg. us-east-1)."""

        return self.regions_by_name[name]

    def get_by_key(self, key: RegionKey) -> Region:
        """Get an AWS region by its key."""

        return self.regions_by_key[key]

    def get_by_id(self, id_: str) -> Region:
        """Get an AWS region by region_id."""

        return self.get_by_key(
            RegionKey(provider_id=self.resource_manager.provider_id, region_id=id_)
        )



from typing import Optional

from firebolt.client.auth.base import AuthRequest
from firebolt.client.auth.request_auth_base import _RequestBasedAuth
from firebolt.utils.token_storage import TokenSecureStorage
from firebolt.utils.urls import AUTH_URL
from firebolt.utils.util import cached_property


class UsernamePassword(_RequestBasedAuth):
    """Username/Password authentication class for Firebolt Database.

    Gets authentication token using
    provided credentials and updates it when it expires.

    Args:
        username (str): Username
        password (str): Password
        use_token_cache (bool): True if token should be cached in filesystem;
            False otherwise

    Attributes:
        username (str): Username
        password (str): Password
    """

    __slots__ = (
        "username",
        "password",
        "_token",
        "_expires",
        "_use_token_cache",
        "_user_agent",
    )

    requires_response_body = True

    def __init__(
        self,
        username: str,
        password: str,
        use_token_cache: bool = True,
    ):
        self.username = username
        self.password = password
        super().__init__(use_token_cache)

    def copy(self) -> "UsernamePassword":
        """Make another auth object with same credentials.

        Returns:
            UsernamePassword: Auth object
        """
        return UsernamePassword(self.username, self.password, self._use_token_cache)

    @cached_property
    def _token_storage(self) -> Optional[TokenSecureStorage]:
        """Token filesystem cache storage.

        This is evaluated lazily, only if caching is enabled

        Returns:
            TokenSecureStorage: Token filesystem cache storage
        """
        return TokenSecureStorage(username=self.username, password=self.password)

    def _make_auth_request(self) -> AuthRequest:
        """Get new token using username and password.

        Yields:
            Request: An http request to get token. Expects Response to be sent back

        Raises:
            AuthenticationError: Error while authenticating with provided credentials
        """
        response = self.request_class(
            "POST",
            AUTH_URL,
            headers={
                "Content-Type": "application/json;charset=UTF-8",
                "User-Agent": self._user_agent,
            },
            json={"username": self.username, "password": self.password},
        )
        return response



from json import JSONDecodeError
from typing import Tuple, Type

from httpx import CookieConflict, HTTPError, InvalidURL, StreamError

DEFAULT_API_URL: str = "api.app.firebolt.io"
_REQUEST_ERRORS: Tuple[Type, ...] = (
    HTTPError,
    InvalidURL,
    CookieConflict,
    StreamError,
    JSONDecodeError,
    KeyError,
    ValueError,
)



from __future__ import annotations

from typing import TYPE_CHECKING

from httpx import URL, Response

from firebolt.utils.urls import DATABASES_URL, ENGINES_URL

if TYPE_CHECKING:
    from firebolt.async_db.connection import Connection


async def is_db_available(connection: Connection, database_name: str) -> bool:
    """
    Verify that the database exists.

    Args:
        connection (firebolt.async_db.connection.Connection)
    """
    resp = await _filter_request(
        connection, DATABASES_URL, {"filter.name_contains": database_name}
    )
    return len(resp.json()["edges"]) > 0


async def is_engine_running(connection: Connection, engine_url: str) -> bool:
    """
    Verify that the engine is running.

    Args:
        connection (firebolt.async_db.connection.Connection): connection.
    """
    # Url is not guaranteed to be of this structure,
    # but for the sake of error checking this is sufficient.
    engine_name = URL(engine_url).host.split(".")[0].replace("-", "_")
    resp = await _filter_request(
        connection,
        ENGINES_URL,
        {
            "filter.name_contains": engine_name,
            "filter.current_status_eq": "ENGINE_STATUS_RUNNING_REVISION_SERVING",
        },
    )
    return len(resp.json()["edges"]) > 0


async def _filter_request(
    connection: Connection, endpoint: str, filters: dict
) -> Response:
    resp = await connection._client.request(
        # Full url overrides the client url, which contains engine as a prefix.
        url=connection.api_endpoint + endpoint,
        method="GET",
        params=filters,
    )
    resp.raise_for_status()
    return resp



import logging

from pydantic import BaseSettings, Field, SecretStr, root_validator

from firebolt.client.auth import Auth

logger = logging.getLogger(__name__)

AUTH_CREDENTIALS_DEPRECATION_MESSAGE = """ Passing connection credentials directly in Settings is deprecated.
 Use Auth object instead.
 Examples:
  >>> from firebolt.client.auth import UsernamePassword
  >>> ...
  >>> settings = Settings(auth=UsernamePassword(username, password), ...)
 or
  >>> from firebolt.client.auth import Token
  >>> ...
  >>> settings = Settings(auth=Token(access_token), ...)"""


class Settings(BaseSettings):
    """Settings for Firebolt SDK.

    Attributes:
        user (Optional[str]): User name
        password (Optional[str]): User password
        access_token (Optional[str]): Access token to use for authentication
            Mutually exclusive with user and password
        account_name (Optional[str]): Account name
            Default user account is used if none provided
        server (Optional[str]): Environment api endpoint (Advanced)
            Default api endpoint is used if none provided
        default_region (str): Default region for provisioning
    """

    auth: Auth = Field(None)
    # Authorization
    user: str = Field(None, env="FIREBOLT_USER")
    password: SecretStr = Field(None, env="FIREBOLT_PASSWORD")
    # Or
    access_token: str = Field(None, env="FIREBOLT_AUTH_TOKEN")

    account_name: str = Field(None, env="FIREBOLT_ACCOUNT")
    server: str = Field(..., env="FIREBOLT_SERVER")
    default_region: str = Field(..., env="FIREBOLT_DEFAULT_REGION")
    use_token_cache: bool = Field(True)

    class Config:
        """Internal pydantic config."""

        env_file = ".env"

    @root_validator
    def mutual_exclusive_with_creds(cls, values: dict) -> dict:
        """Validate that either creds or token is provided.

        Args:
            values (dict): settings initial values

        Returns:
            dict: Validated settings values

        Raises:
            ValueError: Either both or none of credentials and token are provided
        """

        params_present = (
            values.get("user") is not None or values.get("password") is not None,
            values.get("access_token") is not None,
            values.get("auth") is not None,
        )
        if sum(params_present) == 0:
            raise ValueError(
                "Provide at least one of auth, user/password or access_token."
            )
        if sum(params_present) > 1:
            raise ValueError("Provide only one of auth, user/password or access_token")
        if any(values.get(f) for f in ("user", "password", "access_token")):
            logger.warning(AUTH_CREDENTIALS_DEPRECATION_MESSAGE)

        return values



from functools import lru_cache, partial, wraps
from typing import TYPE_CHECKING, Any, Callable, Type, TypeVar

import trio
from httpx import URL

T = TypeVar("T")


def cached_property(func: Callable[..., T]) -> T:
    """cached_property implementation for 3.7 backward compatibility.

    Args:
        func (Callable): Property getter

    Returns:
        T: Property of type, returned by getter
    """
    return property(lru_cache()(func))  # type: ignore


def prune_dict(d: dict) -> dict:
    """Prune items from dictionaries where value is None.

    Args:
        d (dict): Dict to prune

    Returns:
        dict: Pruned dict
    """
    return {k: v for k, v in d.items() if v is not None}


TMix = TypeVar("TMix")


def mixin_for(baseclass: Type[TMix]) -> Type[TMix]:
    """Define mixin with baseclass typehint.

    Should be used as a mixin base class to fix typehints.

    Args:
        baseclass (Type[TMix]): Class which mixin will be made for

    Returns:
        Type[TMix]: Mixin type to inherit from

    Examples:
        ```
        class ReadonlyMixin(mixin_for(BaseClass))):
            ...
        ```

    """
    if TYPE_CHECKING:
        return baseclass
    return object


def fix_url_schema(url: str) -> str:
    """Add schema to URL if it's missing.

    Args:
        url (str): URL to check

    Returns:
        str: URL with schema present

    """
    return url if url.startswith("http") else f"https://{url}"


def async_to_sync(f: Callable) -> Callable:
    """Convert async function to sync.

    Args:
        f (Callable): function to convert

    Returns:
        Callable: regular function, which can be executed synchronously
    """

    @wraps(f)
    def sync(*args: Any, **kwargs: Any) -> Any:
        return trio.run(partial(f, *args, **kwargs))

    return sync


def merge_urls(base: URL, merge: URL) -> URL:
    """Merge a base and merge urls.

    If merge is not a relative url, do nothing

    Args:
        base (URL): Base URL to merge to
        merge (URL): URL to merge

    Returns:
        URL: Resulting URL
    """
    if merge.is_relative_url:
        merge_raw_path = base.raw_path + merge.raw_path.lstrip(b"/")
        return base.copy_with(raw_path=merge_raw_path)
    return merge



from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any, List, Optional, Sequence

from pydantic import Field, PrivateAttr

from firebolt.model import FireboltBaseModel
from firebolt.model.region import RegionKey
from firebolt.service.types import EngineStatusSummary
from firebolt.utils.exception import AttachedEngineInUseError
from firebolt.utils.urls import ACCOUNT_DATABASE_URL

if TYPE_CHECKING:
    from firebolt.model.binding import Binding
    from firebolt.model.engine import Engine
    from firebolt.service.database import DatabaseService

logger = logging.getLogger(__name__)


class DatabaseKey(FireboltBaseModel):
    account_id: str
    database_id: str


class FieldMask(FireboltBaseModel):
    paths: Sequence[str] = Field(alias="paths")


class Database(FireboltBaseModel):
    """
    A Firebolt database.

    Databases belong to a region and have a description,
    but otherwise are not configurable.
    """

    # internal
    _service: DatabaseService = PrivateAttr()

    # required
    name: str = Field(min_length=1, max_length=255, regex=r"^[0-9a-zA-Z_]+$")
    compute_region_key: RegionKey = Field(alias="compute_region_id")

    # optional
    database_key: Optional[DatabaseKey] = Field(alias="id")
    description: Optional[str] = Field(max_length=255)
    emoji: Optional[str] = Field(max_length=255)
    current_status: Optional[str]
    health_status: Optional[str]
    data_size_full: Optional[int]
    data_size_compressed: Optional[int]
    is_system_database: Optional[bool]
    storage_bucket_name: Optional[str]
    create_time: Optional[datetime]
    create_actor: Optional[str]
    last_update_time: Optional[datetime]
    last_update_actor: Optional[str]
    desired_status: Optional[str]

    @classmethod
    def parse_obj_with_service(
        cls, obj: Any, database_service: DatabaseService
    ) -> Database:
        database = cls.parse_obj(obj)
        database._service = database_service
        return database

    @property
    def database_id(self) -> Optional[str]:
        if self.database_key is None:
            return None
        return self.database_key.database_id

    def get_attached_engines(self) -> List[Engine]:
        """Get a list of engines that are attached to this database."""

        return self._service.resource_manager.bindings.get_engines_bound_to_database(  # noqa: E501
            database=self
        )

    def attach_to_engine(
        self, engine: Engine, is_default_engine: bool = False
    ) -> Binding:
        """
        Attach an engine to this database.

        Args:
            engine: The engine to attach.
            is_default_engine:
                Whether this engine should be used as default for this database.
                Only one engine can be set as default for a single database.
                This will overwrite any existing default.
        """

        return self._service.resource_manager.bindings.create(
            engine=engine, database=self, is_default_engine=is_default_engine
        )

    def delete(self) -> Database:
        """
        Delete a database from Firebolt.

        Raises an error if there are any attached engines.
        """

        for engine in self.get_attached_engines():
            if engine.current_status_summary in {
                EngineStatusSummary.ENGINE_STATUS_SUMMARY_STARTING,
                EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPING,
            }:
                raise AttachedEngineInUseError(method_name="delete")

        logger.info(
            f"Deleting Database (database_id={self.database_id}, name={self.name})"
        )
        response = self._service.client.delete(
            url=ACCOUNT_DATABASE_URL.format(
                account_id=self._service.account_id, database_id=self.database_id
            ),
            headers={"Content-type": "application/json"},
        )
        return Database.parse_obj_with_service(
            response.json()["database"], self._service
        )

    def update(self, description: str) -> Database:
        """
        Updates a database description.
        """

        class _DatabaseUpdateRequest(FireboltBaseModel):
            """Helper model for sending Database creation requests."""

            account_id: str
            database: Database
            database_id: str
            update_mask: FieldMask

        self.description = description

        logger.info(
            f"Updating Database (database_id={self.database_id}, "
            f"name={self.name}, description={self.description})"
        )

        payload = _DatabaseUpdateRequest(
            account_id=self._service.account_id,
            database=self,
            database_id=self.database_id,
            update_mask=FieldMask(paths=["description"]),
        ).jsonable_dict(by_alias=True)

        response = self._service.client.patch(
            url=ACCOUNT_DATABASE_URL.format(
                account_id=self._service.account_id, database_id=self.database_id
            ),
            headers={"Content-type": "application/json"},
            json=payload,
        )

        return Database.parse_obj_with_service(
            response.json()["database"], self._service
        )

    def get_default_engine(self) -> Optional[Engine]:
        """
        Returns: default engine of the database, or None if default engine is missing
        """
        rm = self._service.resource_manager
        default_engines = [
            rm.engines.get(binding.engine_id)
            for binding in rm.bindings.get_many(database_id=self.database_id)
            if binding.is_default_engine
        ]

        return None if len(default_engines) == 0 else default_engines[0]



from firebolt.client import Client
from firebolt.common import Settings
from firebolt.service.manager import ResourceManager


class BaseService:
    def __init__(self, resource_manager: ResourceManager):
        self.resource_manager = resource_manager

    @property
    def client(self) -> Client:
        return self.resource_manager.client

    @property
    def account_id(self) -> str:
        return self.resource_manager.account_id

    @property
    def settings(self) -> Settings:
        return self.resource_manager.settings



import inspect
import logging
from importlib import import_module
from pathlib import Path
from platform import python_version, release, system
from sys import modules
from typing import Dict, List, Optional, Tuple

from pydantic import BaseModel

from firebolt import __version__


class ConnectorVersions(BaseModel):
    """
    Verify correct parameter types
    """

    clients: List[Tuple[str, str]]
    drivers: List[Tuple[str, str]]


logger = logging.getLogger(__name__)


# Name, Function, Path, module
CLIENT_MAP = [
    (
        "Airflow",
        "get_conn",
        Path("firebolt_provider/hooks/firebolt.py"),
        "firebolt_provider",
    ),
    (
        "AirbyteDestination",
        "establish_connection",
        Path("destination_firebolt/destination.py"),
        "",
    ),
    (
        "AirbyteDestination",
        "establish_async_connection",
        Path("destination_firebolt/destination.py"),
        "",
    ),
    ("AirbyteSource", "establish_connection", Path("source_firebolt/source.py"), ""),
    (
        "AirbyteSource",
        "establish_async_connection",
        Path("source_firebolt/source.py"),
        "",
    ),
    ("FireboltCLI", "create_connection", Path("firebolt_cli/utils.py"), "firebolt_cli"),
    (
        "DBT",
        "open",
        Path("dbt/adapters/firebolt/connections.py"),
        "dbt.adapters.firebolt",
    ),
    (
        "Superset",
        "",  # connection is created in multiple places
        Path("superset/models/core.py"),
        "",
    ),
    (
        "Redash",
        "run_query",
        Path("redash/query_runner/firebolt.py"),
        "redash",
    ),
    (
        "Prefect",
        "run",
        Path("prefect/tasks/firebolt/firebolt.py"),
        "prefect",
    ),
]

DRIVER_MAP = [
    ("SQLAlchemy", "connect", Path("sqlalchemy/engine/default.py"), "firebolt_db"),
]


def _os_compare(file: Path, expected: Path) -> bool:
    """
    System-independent path comparison.

    Args:
        file: file path to check against
        expected: expected file path

    Returns:
        True if file ends with path
    """
    return file.parts[-len(expected.parts) :] == expected.parts


def get_sdk_properties() -> Tuple[str, str, str, str]:
    """
    Detect Python, OS and SDK versions.

    Returns:
        Python version, SDK version, OS name and "ciso" if imported
    """
    py_version = python_version()
    sdk_version = __version__
    os_version = f"{system()} {release()}"
    ciso = "ciso8601" if "ciso8601" in modules.keys() else ""
    logger.debug(
        "Python %s detected. SDK %s OS %s %s",
        py_version,
        sdk_version,
        os_version,
        ciso,
    )
    return (py_version, sdk_version, os_version, ciso)


def detect_connectors(
    connector_map: List[Tuple[str, str, Path, str]]
) -> Dict[str, str]:
    """
    Detect which connectors are running the code by parsing the stack.
    Exceptions are ignored since this is intended for logging only.
    """
    connectors: Dict[str, str] = {}
    stack = inspect.stack()
    for f in stack:
        try:
            for name, func, path, version_path in connector_map:
                if (not func or f.function == func) and _os_compare(
                    Path(f.filename), path
                ):
                    if version_path:
                        m = import_module(version_path)
                        connectors[name] = m.__version__  # type: ignore
                    else:
                        # Some connectors don't have versions specified
                        connectors[name] = ""
                    # No need to carry on if connector is detected
                    break
        except Exception:
            logger.debug(
                "Failed to extract version from %s in %s", f.function, f.filename
            )
    return connectors


def format_as_user_agent(drivers: Dict[str, str], clients: Dict[str, str]) -> str:
    """
    Return a representation of a stored tracking data as a user-agent header.

    Args:
        connectors: Dictionary of connector to version mappings

    Returns:
        String of the current detected connector stack.
    """
    py, sdk, os, ciso = get_sdk_properties()
    sdk_format = f"PythonSDK/{sdk} (Python {py}; {os}; {ciso})"
    driver_format = "".join(
        [f" {connector}/{version}" for connector, version in drivers.items()]
    )
    client_format = "".join(
        [f"{connector}/{version} " for connector, version in clients.items()]
    )
    return client_format + sdk_format + driver_format


def get_user_agent_header(
    user_drivers: Optional[List[Tuple[str, str]]] = [],
    user_clients: Optional[List[Tuple[str, str]]] = [],
) -> str:
    """
    Return a user agent header with connector stack and system information.

    Args:
        user_drivers(Optional): User-supplied list of tuples of all drivers
            and their versions intended for tracking. Driver is a programmatic
            module that facilitates interaction between a clients and underlying
            database.
        user_clients(Optional): User-supplied list of tuples of all clients
            and their versions intended for tracking. Client is a user-facing
            module or application that allows interaction with the database
            via drivers or directly.

    Returns:
        String representation of a user-agent tracking information
    """
    drivers = detect_connectors(DRIVER_MAP)
    clients = detect_connectors(CLIENT_MAP)
    logger.debug(
        "Detected running with drivers: %s and clients %s ", str(drivers), str(clients)
    )
    # Override auto-detected connectors with info provided manually
    versions = ConnectorVersions(clients=user_clients, drivers=user_drivers)
    for name, version in versions.clients:
        clients[name] = version
    for name, version in versions.drivers:
        drivers[name] = version
    return format_as_user_agent(drivers, clients)



from typing import Dict, List, NamedTuple, Optional

from firebolt.model.instance_type import InstanceType, InstanceTypeKey
from firebolt.model.region import Region
from firebolt.service.base import BaseService
from firebolt.utils.urls import ACCOUNT_INSTANCE_TYPES_URL
from firebolt.utils.util import cached_property


class InstanceTypeLookup(NamedTuple):
    """Helper tuple for looking up instance types by names."""

    region_name: str
    instance_type_name: str


class InstanceTypeService(BaseService):
    @cached_property
    def instance_types(self) -> List[InstanceType]:
        """List of instance types available on Firebolt."""

        response = self.client.get(
            url=ACCOUNT_INSTANCE_TYPES_URL.format(account_id=self.account_id),
            params={"page.first": 5000},
        )
        return [InstanceType.parse_obj(i["node"]) for i in response.json()["edges"]]

    @cached_property
    def instance_types_by_key(self) -> Dict[InstanceTypeKey, InstanceType]:
        """Dict of {InstanceTypeKey to InstanceType}"""

        return {i.key: i for i in self.instance_types}

    @cached_property
    def instance_types_by_name(self) -> Dict[InstanceTypeLookup, InstanceType]:
        """Dict of {InstanceTypeLookup to InstanceType}"""

        return {
            InstanceTypeLookup(
                region_name=self.resource_manager.regions.get_by_id(
                    id_=i.key.region_id
                ).name,
                instance_type_name=i.name,
            ): i
            for i in self.instance_types
        }

    def get_instance_types_per_region(self, region: Region) -> List[InstanceType]:
        """List of instance types available on Firebolt in specified region."""

        response = self.client.get(
            url=ACCOUNT_INSTANCE_TYPES_URL.format(account_id=self.account_id),
            params={"page.first": 5000, "filter.id_region_id_eq": region.key.region_id},
        )

        instance_list = [
            InstanceType.parse_obj(i["node"]) for i in response.json()["edges"]
        ]

        # Filter out instances without storage
        return [
            i
            for i in instance_list
            if i.storage_size_bytes and i.storage_size_bytes != "0"
        ]

    def cheapest_instance_in_region(self, region: Region) -> Optional[InstanceType]:
        # Get only available instances in region
        instance_list = self.get_instance_types_per_region(region)

        if not instance_list:
            return None

        cheapest = min(
            instance_list,
            key=lambda x: x.price_per_hour_cents
            if x.price_per_hour_cents
            else float("Inf"),
        )
        return cheapest

    def get_by_key(self, instance_type_key: InstanceTypeKey) -> InstanceType:
        """Get an instance type by key."""

        return self.instance_types_by_key[instance_type_key]

    def get_by_name(
        self,
        instance_type_name: str,
        region_name: Optional[str] = None,
    ) -> InstanceType:
        """
        Get an instance type by name.

        Args:
            instance_type_name: Name of the instance (eg. "i3.4xlarge")
            region_name:
                Name of the AWS region from which to get the instance.
                If not provided, use the default region name from the client.

        Returns:
            The requested instance type
        """

        # Will raise an error if neither set
        region_name = region_name or self.resource_manager.regions.default_region.name
        return self.instance_types_by_name[
            InstanceTypeLookup(
                region_name=region_name,
                instance_type_name=instance_type_name,
            )
        ]



from time import time
from typing import Generator, Optional

from httpx import Auth as HttpxAuth
from httpx import Request, Response, codes

from firebolt.utils.token_storage import TokenSecureStorage
from firebolt.utils.util import cached_property


class AuthRequest(Request):
    """Class to distinguish auth requests from regular"""


class Auth(HttpxAuth):
    """Base authentication class for Firebolt database.

    Updates all http requests with bearer token authorization header

    Args:
        use_token_cache (bool): True if token should be cached in filesystem;
            False otherwise
    """

    __slots__ = (
        "_token",
        "_expires",
        "_use_token_cache",
    )

    requires_response_body = True
    request_class = AuthRequest

    def __init__(self, use_token_cache: bool = True):
        self._use_token_cache = use_token_cache
        self._token: Optional[str] = self._get_cached_token()
        self._expires: Optional[int] = None

    def copy(self) -> "Auth":
        """Make another auth object with same credentials.

        Returns:
            Auth: Auth object
        """
        return Auth(self._use_token_cache)

    @property
    def token(self) -> Optional[str]:
        """Acquired bearer token.

        Returns:
            Optional[str]: Acquired token
        """
        return self._token

    @property
    def expired(self) -> bool:
        """Check if current token is expired.

        Returns:
            bool: True if expired, False otherwise
        """
        return self._expires is not None and self._expires <= int(time())

    @cached_property
    def _token_storage(self) -> Optional[TokenSecureStorage]:
        """Token filesystem cache storage.

        This is evaluated lazily, only if caching is enabled.

        Returns:
            Optional[TokenSecureStorage]: Token filesystem cache storage if any
        """
        return None

    def _get_cached_token(self) -> Optional[str]:
        """If caching is enabled, get token from filesystem cache.

        If caching is disabled, None is returned.

        Returns:
            Optional[str]: Token if any, and if caching is enabled; None otherwise
        """
        if not self._use_token_cache or not self._token_storage:
            return None
        return self._token_storage.get_cached_token()

    def _cache_token(self) -> None:
        """If caching isenabled, cache token to filesystem."""
        if not self._use_token_cache or not self._token_storage:
            return
        # Only cache if token and expiration are retrieved
        if self._token and self._expires:
            self._token_storage.cache_token(self._token, self._expires)

    def get_new_token_generator(self) -> Generator[Request, Response, None]:
        """Generate requests needed to create a new token session."""

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """Add authorization token to request headers.

        Overrides ``httpx.Auth.auth_flow``

        Args:
            request (Request): Request object to update

        Yields:
            Request: Request required for auth flow
        """
        if not self.token or self.expired:
            yield from self.get_new_token_generator()
            self._cache_token()

        request.headers["Authorization"] = f"Bearer {self.token}"

        response = yield request

        if response.status_code == codes.UNAUTHORIZED:
            yield from self.get_new_token_generator()
            request.headers["Authorization"] = f"Bearer {self.token}"
            yield request



from __future__ import annotations

import logging
import re
import time
from enum import Enum
from functools import wraps
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from aiorwlock import RWLock
from httpx import Response, codes
from pydantic import BaseModel

from firebolt.async_db._types import (
    ColType,
    Column,
    ParameterType,
    RawColType,
    SetParameter,
    parse_type,
    parse_value,
    split_format_sql,
)
from firebolt.async_db.util import is_db_available, is_engine_running
from firebolt.client import AsyncClient
from firebolt.utils.exception import (
    AsyncExecutionUnavailableError,
    CursorClosedError,
    DataError,
    EngineNotRunningError,
    FireboltDatabaseError,
    OperationalError,
    ProgrammingError,
    QueryNotRunError,
)

if TYPE_CHECKING:
    from firebolt.async_db.connection import Connection

logger = logging.getLogger(__name__)


JSON_OUTPUT_FORMAT = "JSONCompact"


class CursorState(Enum):
    NONE = 1
    ERROR = 2
    DONE = 3
    CLOSED = 4


class QueryStatus(Enum):
    """Enumeration of query responses on server-side async queries."""

    RUNNING = 1
    ENDED_SUCCESSFULLY = 2
    ENDED_UNSUCCESSFULLY = 3
    NOT_READY = 4
    STARTED_EXECUTION = 5
    PARSE_ERROR = 6
    CANCELED_EXECUTION = 7
    EXECUTION_ERROR = 8


class Statistics(BaseModel):
    """
    Class for query execution statistics.
    """

    elapsed: float
    rows_read: int
    bytes_read: int
    time_before_execution: float
    time_to_execute: float
    scanned_bytes_cache: Optional[float]
    scanned_bytes_storage: Optional[float]


def check_not_closed(func: Callable) -> Callable:
    """(Decorator) ensure cursor is not closed before calling method."""

    @wraps(func)
    def inner(self: Cursor, *args: Any, **kwargs: Any) -> Any:
        if self.closed:
            raise CursorClosedError(method_name=func.__name__)
        return func(self, *args, **kwargs)

    return inner


def check_query_executed(func: Callable) -> Callable:
    """
    (Decorator) ensure that some query has been executed before
    calling cursor method.
    """

    @wraps(func)
    def inner(self: Cursor, *args: Any, **kwargs: Any) -> Any:
        if self._state == CursorState.NONE:
            raise QueryNotRunError(method_name=func.__name__)
        return func(self, *args, **kwargs)

    return inner


class BaseCursor:
    __slots__ = (
        "connection",
        "_arraysize",
        "_client",
        "_state",
        "_descriptions",
        "_statistics",
        "_rowcount",
        "_rows",
        "_idx",
        "_idx_lock",
        "_row_sets",
        "_next_set_idx",
        "_set_parameters",
        "_query_id",
    )

    default_arraysize = 1

    def __init__(self, client: AsyncClient, connection: Connection):
        self.connection = connection
        self._client = client
        self._arraysize = self.default_arraysize
        # These fields initialized here for type annotations purpose
        self._rows: Optional[List[List[RawColType]]] = None
        self._descriptions: Optional[List[Column]] = None
        self._statistics: Optional[Statistics] = None
        self._row_sets: List[
            Tuple[
                int,
                Optional[List[Column]],
                Optional[Statistics],
                Optional[List[List[RawColType]]],
            ]
        ] = []
        self._set_parameters: Dict[str, Any] = dict()
        self._rowcount = -1
        self._idx = 0
        self._next_set_idx = 0
        self._query_id = ""
        self._reset()

    def __del__(self) -> None:
        self.close()

    @property  # type: ignore
    @check_not_closed
    def description(self) -> Optional[List[Column]]:
        """
        Provides information about a single result row of a query.

        Attributes:
            * ``name``
            * ``type_code``
            * ``display_size``
            * ``internal_size``
            * ``precision``
            * ``scale``
            * ``null_ok``
        """
        return self._descriptions

    @property  # type: ignore
    @check_not_closed
    def statistics(self) -> Optional[Statistics]:
        """Query execution statistics returned by the backend."""
        return self._statistics

    @property  # type: ignore
    @check_not_closed
    def rowcount(self) -> int:
        """The number of rows produced by last query."""
        return self._rowcount

    @property  # type: ignore
    @check_not_closed
    def query_id(self) -> str:
        """The query id of a query executed asynchronously."""
        return self._query_id

    @property
    def arraysize(self) -> int:
        """Default number of rows returned by fetchmany."""
        return self._arraysize

    @arraysize.setter
    def arraysize(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError(
                "Invalid arraysize value type, expected int,"
                f" got {type(value).__name__}"
            )
        self._arraysize = value

    @property
    def closed(self) -> bool:
        """True if connection is closed, False otherwise."""
        return self._state == CursorState.CLOSED

    def close(self) -> None:
        """Terminate an ongoing query (if any) and mark connection as closed."""
        self._state = CursorState.CLOSED
        # remove typecheck skip  after connection is implemented
        self.connection._remove_cursor(self)  # type: ignore

    @check_not_closed
    @check_query_executed
    def nextset(self) -> Optional[bool]:
        """
        Skip to the next available set, discarding any remaining rows
        from the current set.
        Returns True if operation was successful;
        None if there are no more sets to retrive.
        """
        return self._pop_next_set()

    def _pop_next_set(self) -> Optional[bool]:
        """
        Same functionality as .nextset, but doesn't check that query has been executed.
        """
        if self._next_set_idx >= len(self._row_sets):
            return None
        (
            self._rowcount,
            self._descriptions,
            self._statistics,
            self._rows,
        ) = self._row_sets[self._next_set_idx]
        self._idx = 0
        self._next_set_idx += 1
        return True

    def flush_parameters(self) -> None:
        self._set_parameters = dict()

    async def _raise_if_error(self, resp: Response) -> None:
        """Raise a proper error if any"""
        if resp.status_code == codes.INTERNAL_SERVER_ERROR:
            raise OperationalError(
                f"Error executing query:\n{resp.read().decode('utf-8')}"
            )
        if resp.status_code == codes.FORBIDDEN:
            if not await is_db_available(self.connection, self.connection.database):
                raise FireboltDatabaseError(
                    f"Database {self.connection.database} does not exist"
                )
            raise ProgrammingError(resp.read().decode("utf-8"))
        if (
            resp.status_code == codes.SERVICE_UNAVAILABLE
            or resp.status_code == codes.NOT_FOUND
        ):
            if not await is_engine_running(self.connection, self.connection.engine_url):
                raise EngineNotRunningError(
                    f"Firebolt engine {self.connection.engine_url} "
                    "needs to be running to run queries against it."
                )
        resp.raise_for_status()

    def _reset(self) -> None:
        """Clear all data stored from previous query."""
        self._state = CursorState.NONE
        self._rows = None
        self._descriptions = None
        self._statistics = None
        self._rowcount = -1
        self._idx = 0
        self._row_sets = []
        self._next_set_idx = 0
        self._query_id = ""

    def _row_set_from_response(
        self, response: Response
    ) -> Tuple[
        int,
        Optional[List[Column]],
        Optional[Statistics],
        Optional[List[List[RawColType]]],
    ]:
        """Fetch information about executed query from http response."""

        # Empty response is returned for insert query
        if response.headers.get("content-length", "") == "0":
            return (-1, None, None, None)
        try:
            # Skip parsing floats to properly parse them later
            query_data = response.json(parse_float=str)
            rowcount = int(query_data["rows"])
            descriptions = [
                Column(d["name"], parse_type(d["type"]), None, None, None, None, None)
                for d in query_data["meta"]
            ]
            statistics = Statistics(**query_data["statistics"])
            # Parse data during fetch
            rows = query_data["data"]
            return (rowcount, descriptions, statistics, rows)
        except (KeyError, ValueError) as err:
            raise DataError(f"Invalid query data format: {str(err)}")

    def _append_row_set(
        self,
        row_set: Tuple[
            int,
            Optional[List[Column]],
            Optional[Statistics],
            Optional[List[List[RawColType]]],
        ],
    ) -> None:
        """Store information about executed query."""
        self._row_sets.append(row_set)
        if self._next_set_idx == 0:
            # Populate values for first set
            self._pop_next_set()

    async def _api_request(
        self,
        query: Optional[str] = "",
        parameters: Optional[dict[str, Any]] = {},
        path: Optional[str] = "",
        use_set_parameters: Optional[bool] = True,
    ) -> Response:
        """
        Query API, return Response object.

        Args:
            query (str): SQL query
            parameters (Optional[Sequence[ParameterType]]): A sequence of substitution
                parameters. Used to replace '?' placeholders inside a query with
                actual values. Note: In order to "output_format" dict value, it
                    must be an empty string. If no value not specified,
                    JSON_OUTPUT_FORMAT will be used.
            path (str): endpoint suffix, for example "cancel" or "status"
            use_set_parameters: Optional[bool]: Some queries will fail if additional
                set parameters are sent. Setting this to False will allow
                self._set_parameters to be ignored.
        """
        if use_set_parameters:
            parameters = {**(self._set_parameters or {}), **(parameters or {})}
        return await self._client.request(
            url=f"/{path}",
            method="POST",
            params={
                "database": self.connection.database,
                **(parameters or dict()),
            },
            content=query,
        )

    async def _validate_set_parameter(self, parameter: SetParameter) -> None:
        """Validate parameter by executing simple query with it."""
        if parameter.name == "async_execution":
            raise AsyncExecutionUnavailableError(
                "It is not possible to set async_execution using a SET command. "
                "Instead, pass it as an argument to the execute() or "
                "executemany() function."
            )
        resp = await self._api_request("select 1", {parameter.name: parameter.value})
        # Handle invalid set parameter
        if resp.status_code == codes.BAD_REQUEST:
            raise OperationalError(resp.text)
        await self._raise_if_error(resp)

        # set parameter passed validation
        self._set_parameters[parameter.name] = parameter.value

    def _validate_server_side_async_settings(
        self,
        parameters: Sequence[Sequence[ParameterType]],
        queries: List[Union[SetParameter, str]],
        skip_parsing: bool = False,
        async_execution: Optional[bool] = False,
    ) -> None:
        if async_execution and self._set_parameters.get("use_standard_sql", "1") == "0":
            raise AsyncExecutionUnavailableError(
                "It is not possible to execute queries asynchronously if "
                "use_standard_sql=0."
            )
        if parameters and skip_parsing:
            logger.warning(
                "Query formatting parameters are provided but skip_parsing "
                "is specified. They will be ignored."
            )
        non_set_queries = 0
        for query in queries:
            if type(query) is not SetParameter:
                non_set_queries += 1
        if non_set_queries > 1 and async_execution:
            raise AsyncExecutionUnavailableError(
                "It is not possible to execute multi-statement "
                "queries asynchronously."
            )

    async def _do_execute(
        self,
        raw_query: str,
        parameters: Sequence[Sequence[ParameterType]],
        skip_parsing: bool = False,
        async_execution: Optional[bool] = False,
    ) -> None:
        self._reset()
        # Allow users to manually skip parsing for performance improvement.
        queries: List[Union[SetParameter, str]] = (
            [raw_query] if skip_parsing else split_format_sql(raw_query, parameters)
        )
        try:
            for query in queries:

                start_time = time.time()
                # Our CREATE EXTERNAL TABLE queries currently require credentials,
                # so we will skip logging those queries.
                # https://docs.firebolt.io/sql-reference/commands/create-external-table.html
                if isinstance(query, SetParameter) or not re.search(
                    "aws_key_id|credentials", query, flags=re.IGNORECASE
                ):
                    logger.debug(f"Running query: {query}")

                # Define type for mypy
                row_set: Tuple[
                    int,
                    Optional[List[Column]],
                    Optional[Statistics],
                    Optional[List[List[RawColType]]],
                ] = (-1, None, None, None)
                if isinstance(query, SetParameter):
                    await self._validate_set_parameter(query)
                elif async_execution:
                    self._validate_server_side_async_settings(
                        parameters,
                        queries,
                        skip_parsing,
                        async_execution,
                    )
                    response = await self._api_request(
                        query,
                        {
                            "async_execution": 1,
                            "advanced_mode": 1,
                            "output_format": JSON_OUTPUT_FORMAT,
                        },
                    )
                    await self._raise_if_error(response)
                    if response.headers.get("content-length", "") == "0":
                        raise OperationalError("No response to asynchronous query.")
                    resp = response.json()
                    if "query_id" not in resp or resp["query_id"] == "":
                        raise OperationalError(
                            "Invalid response to asynchronous query: missing query_id."
                        )
                    self._query_id = resp["query_id"]
                else:
                    resp = await self._api_request(
                        query, {"output_format": JSON_OUTPUT_FORMAT}
                    )
                    await self._raise_if_error(resp)
                    row_set = self._row_set_from_response(resp)

                self._append_row_set(row_set)

                logger.info(
                    f"Query fetched {self.rowcount} rows in"
                    f" {time.time() - start_time} seconds."
                )

            self._state = CursorState.DONE

        except Exception:
            self._state = CursorState.ERROR
            raise

    @check_not_closed
    async def execute(
        self,
        query: str,
        parameters: Optional[Sequence[ParameterType]] = None,
        skip_parsing: bool = False,
        async_execution: Optional[bool] = False,
    ) -> Union[int, str]:
        """Prepare and execute a database query.

        Supported features:
            Parameterized queries: placeholder characters ('?') are substituted
                with values provided in `parameters`. Values are formatted to
                be properly recognized by database and to exclude SQL injection.
            Multi-statement queries: multiple statements, provided in a single query
                and separated by semicolon, are executed separatelly and sequentially.
                To switch to next statement result, `nextset` method should be used.
            SET statements: to provide additional query execution parameters, execute
                `SET param=value` statement before it. All parameters are stored in
                cursor object until it's closed. They can also be removed with
                `flush_parameters` method call.

        Args:
            query (str): SQL query to execute
            parameters (Optional[Sequence[ParameterType]]): A sequence of substitution
                parameters. Used to replace '?' placeholders inside a query with
                actual values
            skip_parsing (bool): Flag to disable query parsing. This will
                disable parameterized, multi-statement and SET queries,
                while improving performance
            async_execution (bool): flag to determine if query should be asynchronous

        Returns:
            int: Query row count.
        """
        params_list = [parameters] if parameters else []
        await self._do_execute(query, params_list, skip_parsing, async_execution)
        return self.query_id if async_execution else self.rowcount

    @check_not_closed
    async def executemany(
        self,
        query: str,
        parameters_seq: Sequence[Sequence[ParameterType]],
        async_execution: Optional[bool] = False,
    ) -> Union[int, str]:
        """Prepare and execute a database query.

        Supports providing multiple substitution parameter sets, executing them
        as multiple statements sequentially.

        Supported features:
            Parameterized queries: Placeholder characters ('?') are substituted
                with values provided in `parameters`. Values are formatted to
                be properly recognized by database and to exclude SQL injection.
            Multi-statement queries: Multiple statements, provided in a single query
                and separated by semicolon, are executed separately and sequentially.
                To switch to next statement result, use `nextset` method.
            SET statements: To provide additional query execution parameters, execute
                `SET param=value` statement before it. All parameters are stored in
                cursor object until it's closed. They can also be removed with
                `flush_parameters` method call.

        Args:
            query (str): SQL query to execute.
            parameters_seq (Sequence[Sequence[ParameterType]]): A sequence of
               substitution parameter sets. Used to replace '?' placeholders inside a
               query with actual values from each set in a sequence. Resulting queries
               for each subset are executed sequentially.
            async_execution (bool): flag to determine if query should be asynchronous

        Returns:
            int|str: Query row count for synchronous execution of queries,
            query ID string for asynchronous execution.
        """
        await self._do_execute(query, parameters_seq, async_execution=async_execution)
        if async_execution:
            return self.query_id
        else:
            return self.rowcount

    def _parse_row(self, row: List[RawColType]) -> List[ColType]:
        """Parse a single data row based on query column types."""
        assert len(row) == len(self.description)
        return [
            parse_value(col, self.description[i].type_code) for i, col in enumerate(row)
        ]

    def _get_next_range(self, size: int) -> Tuple[int, int]:
        """
        Return range of next rows of size (if possible),
        and update _idx to point to the end of this range
        """

        if self._rows is None:
            # No elements to take
            raise DataError("no rows to fetch")

        left = self._idx
        right = min(self._idx + size, len(self._rows))
        self._idx = right
        return left, right

    @check_not_closed
    @check_query_executed
    def fetchone(self) -> Optional[List[ColType]]:
        """Fetch the next row of a query result set."""
        left, right = self._get_next_range(1)
        if left == right:
            # We are out of elements
            return None
        assert self._rows is not None
        return self._parse_row(self._rows[left])

    @check_not_closed
    @check_query_executed
    def fetchmany(self, size: Optional[int] = None) -> List[List[ColType]]:
        """
        Fetch the next set of rows of a query result;
        cursor.arraysize is default size.
        """
        size = size if size is not None else self.arraysize
        left, right = self._get_next_range(size)
        assert self._rows is not None
        rows = self._rows[left:right]
        return [self._parse_row(row) for row in rows]

    @check_not_closed
    @check_query_executed
    def fetchall(self) -> List[List[ColType]]:
        """Fetch all remaining rows of a query result."""
        left, right = self._get_next_range(self.rowcount)
        assert self._rows is not None
        rows = self._rows[left:right]
        return [self._parse_row(row) for row in rows]

    @check_not_closed
    def setinputsizes(self, sizes: List[int]) -> None:
        """Predefine memory areas for query parameters (does nothing)."""

    @check_not_closed
    def setoutputsize(self, size: int, column: Optional[int] = None) -> None:
        """Set a column buffer size for fetches of large columns (does nothing)."""

    @check_not_closed
    async def cancel(self, query_id: str) -> None:
        """Cancel a server-side async query."""
        await self._api_request(
            parameters={"query_id": query_id},
            path="cancel",
            use_set_parameters=False,
        )

    @check_not_closed
    async def get_status(self, query_id: str) -> QueryStatus:
        """Get status of a server-side async query. Return the state of the query."""
        try:
            resp = await self._api_request(
                # output_format must be empty for status to work correctly.
                # And set parameters will cause 400 errors.
                parameters={"query_id": query_id},
                path="status",
                use_set_parameters=False,
            )
            if resp.status_code == codes.BAD_REQUEST:
                raise OperationalError(
                    f"Asynchronous query {query_id} status check failed: "
                    f"{resp.status_code}."
                )
            resp_json = resp.json()
            if "status" not in resp_json:
                raise OperationalError(
                    f"Invalid response to asynchronous query: missing status."
                )
        except Exception:
            self._state = CursorState.ERROR
            raise
        # Remember that query_id might be empty.
        if resp_json["status"] == "":
            return QueryStatus.NOT_READY
        return QueryStatus[resp_json["status"]]

    # Context manager support
    @check_not_closed
    def __enter__(self) -> BaseCursor:
        return self

    def __exit__(
        self, exc_type: type, exc_val: Exception, exc_tb: TracebackType
    ) -> None:
        self.close()


class Cursor(BaseCursor):
    """
    Executes async queries to Firebolt Database.
    Should not be created directly;
    use :py:func:`connection.cursor <firebolt.async_db.connection.Connection>`

    Args:
        description: Information about a single result row.
        rowcount: The number of rows produced by last query.
        closed: True if connection is closed; False otherwise.
        arraysize: Read/Write, specifies the number of rows to fetch at a time
            with the :py:func:`fetchmany` method.

    """

    __slots__ = BaseCursor.__slots__ + ("_async_query_lock",)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._async_query_lock = RWLock()
        super().__init__(*args, **kwargs)

    @wraps(BaseCursor.execute)
    async def execute(
        self,
        query: str,
        parameters: Optional[Sequence[ParameterType]] = None,
        skip_parsing: bool = False,
        async_execution: Optional[bool] = False,
    ) -> Union[int, str]:
        async with self._async_query_lock.writer:
            return await super().execute(
                query, parameters, skip_parsing, async_execution
            )

    @wraps(BaseCursor.executemany)
    async def executemany(
        self,
        query: str,
        parameters_seq: Sequence[Sequence[ParameterType]],
        async_execution: Optional[bool] = False,
    ) -> int:
        """
        Prepare and execute a database query against all parameter
        sequences provided.
        """
        async with self._async_query_lock.writer:
            return await super().executemany(query, parameters_seq, async_execution)

    @wraps(BaseCursor.fetchone)
    async def fetchone(self) -> Optional[List[ColType]]:
        async with self._async_query_lock.reader:
            """Fetch the next row of a query result set."""
            return super().fetchone()

    @wraps(BaseCursor.fetchmany)
    async def fetchmany(self, size: Optional[int] = None) -> List[List[ColType]]:
        async with self._async_query_lock.reader:
            """
            Fetch the next set of rows of a query result;
            size is cursor.arraysize by default.
            """
            return super().fetchmany(size)

    @wraps(BaseCursor.fetchall)
    async def fetchall(self) -> List[List[ColType]]:
        async with self._async_query_lock.reader:
            """Fetch all remaining rows of a query result."""
            return super().fetchall()

    @wraps(BaseCursor.nextset)
    async def nextset(self) -> None:
        async with self._async_query_lock.reader:
            return super().nextset()

    # Iteration support
    @check_not_closed
    @check_query_executed
    def __aiter__(self) -> Cursor:
        return self

    @check_not_closed
    @check_query_executed
    async def __anext__(self) -> List[ColType]:
        row = await self.fetchone()
        if row is None:
            raise StopAsyncIteration
        return row



from __future__ import annotations

from functools import wraps
from types import TracebackType
from typing import Any
from warnings import warn

from readerwriterlock.rwlock import RWLockWrite

from firebolt.async_db.connection import BaseConnection as AsyncBaseConnection
from firebolt.async_db.connection import async_connect_factory
from firebolt.db.cursor import Cursor
from firebolt.utils.exception import ConnectionClosedError
from firebolt.utils.util import async_to_sync


class Connection(AsyncBaseConnection):
    """
    Firebolt database connection class. Implements PEP-249.

    Args:

        engine_url: Firebolt database engine REST API url
        database: Firebolt database name
        username: Firebolt account username
        password: Firebolt account password
        api_endpoint: Optional. Firebolt API endpoint. Used for authentication.

    Note:
        Firebolt currenly doesn't support transactions so commit and rollback methods
        are not implemented.
    """

    __slots__ = AsyncBaseConnection.__slots__ + ("_closing_lock",)

    cursor_class = Cursor

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # Holding this lock for write means that connection is closing itself.
        # cursor() should hold this lock for read to read/write state
        self._closing_lock = RWLockWrite()

    def cursor(self) -> Cursor:
        with self._closing_lock.gen_rlock():
            c = super()._cursor()
            assert isinstance(c, Cursor)  # typecheck
            return c

    @wraps(AsyncBaseConnection._aclose)
    def close(self) -> None:
        with self._closing_lock.gen_wlock():
            async_to_sync(self._aclose)()

    # Context manager support
    def __enter__(self) -> Connection:
        if self.closed:
            raise ConnectionClosedError("Connection is already closed.")
        return self

    def __exit__(
        self, exc_type: type, exc_val: Exception, exc_tb: TracebackType
    ) -> None:
        self.close()

    def __del__(self) -> None:
        if not self.closed:
            warn(f"Unclosed {self!r}", UserWarning)


connect = async_to_sync(async_connect_factory(Connection))



from __future__ import annotations

import logging
import socket
from json import JSONDecodeError
from types import TracebackType
from typing import Any, Callable, Dict, List, Optional, Type

from httpcore.backends.auto import AutoBackend
from httpcore.backends.base import AsyncNetworkStream
from httpx import AsyncHTTPTransport, HTTPStatusError, RequestError, Timeout

from firebolt.async_db.cursor import BaseCursor, Cursor
from firebolt.client import DEFAULT_API_URL, AsyncClient
from firebolt.client.auth import Auth, Token, UsernamePassword
from firebolt.utils.exception import (
    ConfigurationError,
    ConnectionClosedError,
    FireboltEngineError,
    InterfaceError,
)
from firebolt.utils.urls import (
    ACCOUNT_ENGINE_ID_BY_NAME_URL,
    ACCOUNT_ENGINE_URL,
    ACCOUNT_ENGINE_URL_BY_DATABASE_NAME,
)
from firebolt.utils.usage_tracker import get_user_agent_header
from firebolt.utils.util import fix_url_schema

DEFAULT_TIMEOUT_SECONDS: int = 60
KEEPALIVE_FLAG: int = 1
KEEPIDLE_RATE: int = 60  # seconds
AUTH_CREDENTIALS_DEPRECATION_MESSAGE = """ Passing connection credentials
 directly to the `connect` function is deprecated.
 Pass the `Auth` object instead.
 Examples:
  >>> from firebolt.client.auth import UsernamePassword
  >>> ...
  >>> connect(auth=UsernamePassword(username, password), ...)
 or
  >>> from firebolt.client.auth import Token
  >>> ...
  >>> connect(auth=Token(access_token), ...)"""

logger = logging.getLogger(__name__)


async def _resolve_engine_url(
    engine_name: str,
    auth: Auth,
    api_endpoint: str,
    account_name: Optional[str] = None,
) -> str:
    async with AsyncClient(
        auth=auth,
        base_url=api_endpoint,
        account_name=account_name,
        api_endpoint=api_endpoint,
        timeout=Timeout(DEFAULT_TIMEOUT_SECONDS),
    ) as client:
        account_id = await client.account_id
        url = ACCOUNT_ENGINE_ID_BY_NAME_URL.format(account_id=account_id)
        try:
            response = await client.get(
                url=url,
                params={"engine_name": engine_name},
            )
            response.raise_for_status()
            engine_id = response.json()["engine_id"]["engine_id"]
            url = ACCOUNT_ENGINE_URL.format(account_id=account_id, engine_id=engine_id)
            response = await client.get(url=url)
            response.raise_for_status()
            return response.json()["engine"]["endpoint"]
        except HTTPStatusError as e:
            # Engine error would be 404.
            if e.response.status_code != 404:
                raise InterfaceError(
                    f"Error {e.__class__.__name__}: Unable to retrieve engine "
                    f"endpoint {url}."
                )
            # Once this is point is reached we've already authenticated with
            # the backend so it's safe to assume the cause of the error is
            # missing engine.
            raise FireboltEngineError(f"Firebolt engine {engine_name} does not exist.")
        except (JSONDecodeError, RequestError, RuntimeError, HTTPStatusError) as e:
            raise InterfaceError(
                f"Error {e.__class__.__name__}: "
                f"Unable to retrieve engine endpoint {url}."
            )


async def _get_database_default_engine_url(
    database: str,
    auth: Auth,
    api_endpoint: str,
    account_name: Optional[str] = None,
) -> str:
    async with AsyncClient(
        auth=auth,
        base_url=api_endpoint,
        account_name=account_name,
        api_endpoint=api_endpoint,
        timeout=Timeout(DEFAULT_TIMEOUT_SECONDS),
    ) as client:
        try:
            account_id = await client.account_id
            response = await client.get(
                url=ACCOUNT_ENGINE_URL_BY_DATABASE_NAME.format(account_id=account_id),
                params={"database_name": database},
            )
            response.raise_for_status()
            return response.json()["engine_url"]
        except (
            JSONDecodeError,
            RequestError,
            RuntimeError,
            HTTPStatusError,
            KeyError,
        ) as e:
            raise InterfaceError(f"Unable to retrieve default engine endpoint: {e}.")


def _validate_engine_name_and_url(
    engine_name: Optional[str], engine_url: Optional[str]
) -> None:
    if engine_name and engine_url:
        raise ConfigurationError(
            "Both engine_name and engine_url are provided. Provide only one to connect."
        )


def _get_auth(
    username: Optional[str],
    password: Optional[str],
    access_token: Optional[str],
    use_token_cache: bool,
) -> Auth:
    """Create `Auth` class based on provided credentials.

    If `access_token` is provided, it's used for `Auth` creation.
    Otherwise, username/password are used.

    Returns:
        Auth: `auth object`

    Raises:
        `ConfigurationError`: Invalid combination of credentials provided

    """
    if not access_token:
        if not username or not password:
            raise ConfigurationError(
                "Neither username/password nor access_token are provided. Provide one"
                " to authenticate."
            )
        return UsernamePassword(username, password, use_token_cache)
    if username or password:
        raise ConfigurationError(
            "Username/password and access_token are both provided. Provide only one"
            " to authenticate."
        )
    return Token(access_token)


def async_connect_factory(connection_class: Type) -> Callable:
    async def connect_inner(
        database: str = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        access_token: Optional[str] = None,
        auth: Auth = None,
        engine_name: Optional[str] = None,
        engine_url: Optional[str] = None,
        account_name: Optional[str] = None,
        api_endpoint: str = DEFAULT_API_URL,
        use_token_cache: bool = True,
        additional_parameters: Dict[str, Any] = {},
    ) -> Connection:
        """Connect to Firebolt database.

        Args:
            `database` (str): Name of the database to connect
            `username` (Optional[str]): User name to use for authentication (Deprecated)
            `password` (Optional[str]): Password to use for authentication (Deprecated)
            `access_token` (Optional[str]): Authentication token to use instead of
                                          credentials (Deprecated)
            `auth` (Auth)L Authentication object.
            `engine_name` (Optional[str]): Name of the engine to connect to
            `engine_url` (Optional[str]): The engine endpoint to use
            `account_name` (Optional[str]): For customers with multiple accounts;
                                          if none, default is used
            `api_endpoint` (str): Firebolt API endpoint. Used for authentication
            `use_token_cache` (bool): Cached authentication token in filesystem
                                    Default: True
            `additional_parameters` (Optional[Dict]): Dictionary of less widely-used
                                    arguments for connection

        Note:
            Providing both `engine_name` and `engine_url` will result in an error

        """
        # These parameters are optional in function signature
        # but are required to connect.
        # PEP 249 recommends making them kwargs.
        if not database:
            raise ConfigurationError("database name is required to connect.")

        _validate_engine_name_and_url(engine_name, engine_url)

        if not auth:
            if any([username, password, access_token, api_endpoint, use_token_cache]):
                logger.warning(AUTH_CREDENTIALS_DEPRECATION_MESSAGE)
                auth = _get_auth(username, password, access_token, use_token_cache)
            else:
                raise ConfigurationError("No authentication provided.")
        api_endpoint = fix_url_schema(api_endpoint)

        # Mypy checks, this should never happen
        assert database is not None

        if not engine_name and not engine_url:
            engine_url = await _get_database_default_engine_url(
                database=database,
                auth=auth,
                account_name=account_name,
                api_endpoint=api_endpoint,
            )

        elif engine_name:
            engine_url = await _resolve_engine_url(
                engine_name=engine_name,
                auth=auth,
                account_name=account_name,
                api_endpoint=api_endpoint,
            )
        elif account_name:
            # In above if branches account name is validated since it's used to
            # resolve or get an engine url.
            # We need to manually validate account_name if none of the above
            # cases are triggered.
            async with AsyncClient(
                auth=auth,
                base_url=api_endpoint,
                account_name=account_name,
                api_endpoint=api_endpoint,
            ) as client:
                await client.account_id

        assert engine_url is not None

        engine_url = fix_url_schema(engine_url)
        return connection_class(
            engine_url, database, auth, api_endpoint, additional_parameters
        )

    return connect_inner


class OverriddenHttpBackend(AutoBackend):
    """
    `OverriddenHttpBackend` is a short-term solution for the TCP
    connection idle timeout issue described in the following article:
    https://docs.aws.amazon.com/elasticloadbalancing/latest/network/network-load-balancers.html#connection-idle-timeout
    Since httpx creates a connection right before executing a request, the
    backend must be overridden to set the socket to `KEEPALIVE`
    and `KEEPIDLE` settings.
    """

    async def connect_tcp(
        self,
        host: str,
        port: int,
        timeout: Optional[float] = None,
        local_address: Optional[str] = None,
    ) -> AsyncNetworkStream:
        stream = await super().connect_tcp(
            host, port, timeout=timeout, local_address=local_address
        )
        # Enable keepalive
        stream.get_extra_info("socket").setsockopt(
            socket.SOL_SOCKET, socket.SO_KEEPALIVE, KEEPALIVE_FLAG
        )
        # MacOS does not have TCP_KEEPIDLE
        if hasattr(socket, "TCP_KEEPIDLE"):
            keepidle = socket.TCP_KEEPIDLE
        else:
            keepidle = 0x10  # TCP_KEEPALIVE on mac

        # Set keepalive to 60 seconds
        stream.get_extra_info("socket").setsockopt(
            socket.IPPROTO_TCP, keepidle, KEEPIDLE_RATE
        )
        return stream


class BaseConnection:
    client_class: type
    cursor_class: type
    __slots__ = (
        "_client",
        "_cursors",
        "database",
        "engine_url",
        "api_endpoint",
        "_is_closed",
    )

    def __init__(
        self,
        engine_url: str,
        database: str,
        auth: Auth,
        api_endpoint: str = DEFAULT_API_URL,
        additional_parameters: Dict[str, Any] = {},
    ):
        # Override tcp keepalive settings for connection
        transport = AsyncHTTPTransport()
        transport._pool._network_backend = OverriddenHttpBackend()
        user_drivers = additional_parameters.get("user_drivers", [])
        user_clients = additional_parameters.get("user_clients", [])
        self._client = AsyncClient(
            auth=auth,
            base_url=engine_url,
            api_endpoint=api_endpoint,
            timeout=Timeout(DEFAULT_TIMEOUT_SECONDS, read=None),
            transport=transport,
            headers={"User-Agent": get_user_agent_header(user_drivers, user_clients)},
        )
        self.api_endpoint = api_endpoint
        self.engine_url = engine_url
        self.database = database
        self._cursors: List[BaseCursor] = []
        self._is_closed = False

    def _cursor(self, **kwargs: Any) -> BaseCursor:
        """
        Create new cursor object.
        """

        if self.closed:
            raise ConnectionClosedError("Unable to create cursor: connection closed.")

        c = self.cursor_class(self._client, self, **kwargs)
        self._cursors.append(c)
        return c

    async def _aclose(self) -> None:
        """Close connection and all underlying cursors."""
        if self.closed:
            return

        # self._cursors is going to be changed during closing cursors
        # after this point no cursors would be added to _cursors, only removed since
        # closing lock is held, and later connection will be marked as closed
        cursors = self._cursors[:]
        for c in cursors:
            # Here c can already be closed by another thread,
            # but it shouldn't raise an error in this case
            c.close()
        await self._client.aclose()
        self._is_closed = True

    @property
    def closed(self) -> bool:
        """`True` if connection is closed; `False` otherwise."""
        return self._is_closed

    def _remove_cursor(self, cursor: Cursor) -> None:
        # This way it's atomic
        try:
            self._cursors.remove(cursor)
        except ValueError:
            pass

    def commit(self) -> None:
        """Does nothing since Firebolt doesn't have transactions."""

        if self.closed:
            raise ConnectionClosedError("Unable to commit: Connection closed.")


class Connection(BaseConnection):
    """
    Firebolt asynchronous database connection class. Implements `PEP 249`_.

    Args:
        `engine_url`: Firebolt database engine REST API url
        `database`: Firebolt database name
        `username`: Firebolt account username
        `password`: Firebolt account password
        `api_endpoint`: Optional. Firebolt API endpoint used for authentication
        `connector_versions`: Optional. Tuple of connector name and version, or
            a list of tuples of your connector stack. Useful for tracking custom
            connector usage.

    Note:
        Firebolt does not support transactions,
        so commit and rollback methods are not implemented.

    .. _PEP 249:
        https://www.python.org/dev/peps/pep-0249/

    """

    cursor_class = Cursor

    aclose = BaseConnection._aclose

    def cursor(self) -> Cursor:
        c = super()._cursor()
        assert isinstance(c, Cursor)  # typecheck
        return c

    # Context manager support
    async def __aenter__(self) -> Connection:
        if self.closed:
            raise ConnectionClosedError("Connection is already closed.")
        return self

    async def __aexit__(
        self, exc_type: type, exc_val: Exception, exc_tb: TracebackType
    ) -> None:
        await self._aclose()


connect = async_connect_factory(Connection)



from datetime import datetime
from typing import Optional

from pydantic import Field

from firebolt.model import FireboltBaseModel


class Provider(FireboltBaseModel, frozen=True):  # type: ignore
    provider_id: str = Field(alias="id")
    name: str

    # optional
    create_time: Optional[datetime]
    display_name: Optional[str]
    last_update_time: Optional[datetime]



from json import JSONDecodeError
from logging import getLogger

from httpx import HTTPStatusError, Request, RequestError, Response

logger = getLogger(__name__)


def log_request(request: Request) -> None:
    """Log HTTP requests.

    Hook for an HTTP client

    Args:
        request (Request): Request to log
    """
    logger.debug(
        "Request event hook: %s %s - Waiting for response", request.method, request.url
    )


def log_response(response: Response) -> None:
    """Log HTTP response.

    Hook for an HTTP client

    Args:
        response (Response): Response to log
    """
    request = response.request
    logger.debug(
        "Response event hook: %s %s - Status %s",
        request.method,
        request.url,
        response.status_code,
    )


def raise_on_4xx_5xx(response: Response) -> None:
    """Raise an error on HTTP response with error return code.

    Hook for an HTTP client.
    If an error message is found, raise as an ApiError.

    Args:
        response (Response): Response to check for error code

    Raises:
        RequestError: Error during performing request
        RuntimeError: Error processing request on server
        HTTPStatusError: HTTP error
    """
    try:
        response.raise_for_status()
    except RequestError as exc:
        logger.debug("An error occurred while requesting %s.", exc.request.url)
        raise exc
    except HTTPStatusError as exc:
        response.read()  # without this, you can get a ResponseNotRead error
        try:
            parsed_response = exc.response.json()
        except JSONDecodeError:
            parsed_response = {"_raw": exc.response.text}
        debug_message = (
            f"Error response {exc.response.status_code} "
            f"while requesting {exc.request.url!r}. "
            f"Response: {parsed_response}. "
        )
        if "message" in parsed_response:
            error_message = parsed_response["message"]
            logger.debug("%s Message: %s", debug_message, error_message)
            raise RuntimeError(error_message) from exc
        logger.debug(debug_message)
        raise exc
